# Copyright 2005 by
#           Atte Timonen
#           Tarmo Toikkanen
#           Teemu Leinonen
#           Lily Diaz
#           Janne Pietarila
#
# Code snippets used from Fle3 (fle3.uiah.fi),
# credits go to the Fle3 Team.
#
# This file is part of the ImaNote product.
#
# ImaNote is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# ImaNote is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ImaNote; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# Python
import re
import time
import string
# Zope
import DateTime
from AccessControl import ClassSecurityInfo
from Globals import InitializeClass
from Acquisition import Implicit
from Globals import Persistent
import OFS
from AccessControl.Role import RoleManager
from Products.ZCatalog.CatalogAwareness import CatalogAware
from common import *

class Annotation(Implicit, OFS.SimpleItem.Item, Persistent, RoleManager,
                 CatalogAware):
    """ Abstract base class. """
    meta_type="Annotation"
    security = ClassSecurityInfo()
    security.declareObjectPublic()
    def __init__(self, left, top, right, bottom, id):
        """ Constructs annotation. """
        self._coords=(left, top, right, bottom)
        self._text=self._url=self._title=self._urlName=self._description=\
            self._author=self._urlAuthor=''
        self._public=True

        self._date=time.time()
        self.id=id
        self.default_catalog = 'catalog_annotations'
        self.annotationUpdated()
        self._keywords=''
        self._icon=None
        self._mimeType=None
        self._duration=None

    def manage_afterAdd(self, item, container):
      CatalogAware.manage_afterAdd(self, item, container)
      self.manage_setLocalRoles(self.getUser(), ('Owner',))
      

    security.declarePrivate('annotationUpdated')
    def annotationUpdated(self):
      self._center_x = int((self._coords[0] + self._coords[2]) / 2.0)
      self._center_y = int((self._coords[1] + self._coords[3]) / 2.0)
      self._width = int(self._coords[2] - self._coords[0])
      self._height = int(self._coords[3] - self._coords[1])
      self._date=time.time()
      self._p_changed=1
      self.reindex_object()
      
    security.declareProtected(PERM_VIEW, 'get_position')
    def getPosition(self, window_width, window_height, magnification_step):
        """Return a tuple (center_x, center_y, zoom_level) where center_x and
        center_y is a center point of annotation area. Variable zoom_level is
        selected so that whole annotation area will fit to window with
        given size when annotation area is centered in that window."""

        # Some margin so that annotation area will not cover the whole window.
        margin = 10
        # Don't use margin if the window size is too small
        if window_width <= 3*margin or window_height <= 3*margin: margin = 0
        zoom_level = 0
        m = pow(magnification_step, zoom_level)
        while (self._width / m) > (window_width - 2*margin) or \
              (self._height /m) > (window_width - 2*margin):
            zoom_level += 1
            m = pow(magnification_step, zoom_level)
        return (self._center_x, self._center_y, zoom_level)
      
  
    security.declarePrivate('move')
    def move(self, xMove, yMove):
      """Move annotation by xMove and yMove"""
      self._coords=(self._coords[0]+xMove, self._coords[1]+yMove,\
          self._coords[2]+xMove, self._coords[3]+yMove)
      self.annotationUpdated()

    security.declarePrivate('moveTo')
    def moveTo(self, xMove, yMove):
      """Move annotation upper left corner to xMove yMove"""
      self._coords=(xMove,yMove,xMove+self._width,yMove+self._height)
      self.annotationUpdated()

    security.declarePrivate('setDimensions')
    def setDimensions(self, width, height):
      """Set new width and height"""
      self._coords=(self._coords[0], self._coords[1], \
          self._coords[0]+width, self._coords[1]+height)
      self.annotationUpdated()
    
    security.declarePrivate('editData')
    def editData(self,coords='', title='', author='', url='', urlName='',\
        text='', description='', public='', keywords='', urlAuthor=''):
      """Edit annotation data"""
      oldurl=self._url
      for field in ['coords','title', 'author', 'url', 'urlName', 'text',\
          'description', 'urlAuthor']:
            exec('if %s==None: self._%s=\'\'' % (field, field))
            exec('if %s: self._%s=%s' % (field, field, field))
      for box in ['public',]:
        exec('self._%s=%s' % (box, box))
      if oldurl!=self._url:
        try:
          url=HeaderURLopener.open(HeaderURLopener(), self._url)
          self._mimeType=url.info().gettype()
        except IOError:
          self._mimeType='text/plain'
      if keywords==None:
        self._keywords=''
      elif keywords:        
        self._keywords=parseKeywords(keywords)
      self.annotationUpdated()

    security.declareProtected(PERM_VIEW, 'getMimeType')
    def getMimeType(self):
      """Return mime type of the link in this annotation"""
      return self._mimeType
    
    security.declareProtected(PERM_VIEW, 'getMediaType')
    def getMediaType(self):
      """Return media type of the link in this annotation"""
      return self._mimeType.split('/')[0]

    security.declareProtected(PERM_VIEW, 'getIcon')
    def getIcon(self):
      """Return icon of this annotation"""
      return self._icon
    
    security.declareProtected(PERM_VIEW, 'getCoordinates')
    def getCoordinates(self):
        """Return coordinates (4-tuple)."""
        return self._coords

    security.declareProtected(PERM_VIEW, 'getTitle')
    def getTitle(self):
      """Return hotspot title"""
      return self._title

    security.declareProtected(PERM_VIEW, 'getAuthor')
    def getAuthor(self):
      """Return annotation author"""
      return self._author

    security.declareProtected(PERM_VIEW, 'getUrlAuthor')
    def getUrlAuthor(self):
      """Return annotation URL author"""
      return self._urlAuthor

    security.declareProtected(PERM_VIEW, 'getUrl')
    def getUrl(self):
      """Return hotspot url"""
      return self._url
    
    security.declareProtected(PERM_VIEW, 'getText')
    def getText(self):
      """Return hotspot text"""
      return self._text
    
    security.declareProtected(PERM_VIEW, 'getUrlName')
    def getUrlName(self):
      """Return hotspot urlName or url if urlName is not specified"""
      if not self._urlName: return self._url
      else: return self._urlName

    security.declareProtected(PERM_VIEW, 'getDescription')
    def getDescription(self):
      """Return hotspot description"""
      return self._description

    security.declareProtected(PERM_VIEW, 'getData')
    def getData(self):
      """Return title, text, url, url name and url description concatenated"""
      return string.join([self._title, self._text, self._url, self._urlName,
          self._description, self._urlAuthor], " ")
      
    security.declareProtected(PERM_VIEW, 'getAllData')
    def getAllData(self):
      """Return title, text, url, url name and url description concatenated"""
      return string.join([self._title, self._text, self._url, self._urlName,
          self._description, self._author, self._urlAuthor], " ")
    
    security.declareProtected(PERM_VIEW, 'getDate')
    def getDate(self):
      """Return hotspot creation/edit date"""
      return self._date

    security.declareProtected(PERM_VIEW, 'getDateTime')
    def getDateTime(self):
      """Return hotspot creation/edit date as DateTime-object"""
      return DateTime.DateTime(self._date)

    security.declareProtected(PERM_VIEW, 'getDateAsCommonF')
    def getDateAsCommonF(self):
      return DateTime.DateTime(self._date).fCommon()

    security.declareProtected(PERM_VIEW, 'getSize')
    def getSize(self):
        """Return size of the area that Annotation covers."""
        return self._width*self._height

    security.declareProtected(PERM_VIEW, 'getKeywords')
    def getKeywords(self):
      """Return keywords"""
      return self._keywords
    
    import re
    security.declareProtected(PERM_VIEW, 'getKeywordList')
    def getKeywordList(self):
      """Return keywords as a nicely formatted string"""
      temp=', '.join(self._keywords.split(' '))
      p=re.compile('([^_\s]*_[^\s]*),')
      temp=p.split(temp)
      for i in range(len(temp)):
        temp[i]=temp[i].strip()
        if temp[i].find('_') is not -1:
          temp[i]=temp[i].replace('_', ' ')
          temp[i]='"'+temp[i]+'",'
      temp=' '.join(temp)
      temp=temp.strip()
      if len(temp)>0 and temp[-1]==',':
        temp=temp[0:-1]
      return temp

    security.declareProtected(PERM_VIEW, 'getKeywordList2')
    def getKeywordList2(self):
      """Same as getKeywrodList, but without commas 
      (the format in which the user inputs the data"""
      return self.getKeywordList().replace(',', '')
    
    security.declareProtected(PERM_VIEW, 'isPublic')
    def isPublic(self):
      """Return whether this annotation is public"""
      #FIXME :it's quite easy to search private annotations, 
      #only ui currently prevents it!
      return self._public

    def doesCover(self, x, y):
        """Does given point belong to Annotation's rectangle?"""
        
        return x >= self._coords[0] and x <= self._coords[2] and \
               y >= self._coords[1]  and y <= self._coords[3]
      
    security.declarePublic('mayEdit')
    def mayEdit(self):
      """Return whether current user can edit this annotation"""
      roles=getLocalRoles(self, self.getUser())
      return 'Manager' in roles or 'Owner' in roles
      
    security.declarePublic('mayView')
    def mayView(self):
      """Return whether current user can view this annotation"""
      roles=getLocalRoles(self, self.getUser())
      return self.isPublic() or 'Owner' in roles

    security.declareProtected(PERM_VIEW, 'getLocateUrl')
    def getLocateUrl(self):
      url=self.absolute_url().split('/annotations')
      urlEnd=url[1].split('/'+self.id)[0]
      return url[0]+urlEnd+'/locateAnnotation?id='+self.id
  
def parseKeywords(keywords):
  """Parse keywords so that they can be searched
  (change '"foo bar" baz' to 'foo_bar baz')"""
  kList=keywords.split('"')
  for i in range(len(kList)):
    if (i%2==1): kList[i]=kList[i].replace(' ','_')
  keywords=' '.join(kList)
  while keywords.find('  ') is not -1:
    keywords=keywords.replace('  ',' ')
  return keywords


InitializeClass(Annotation)
