# Copyright 2005 by
#           Atte Timonen
#           Tarmo Toikkanen
#           Teemu Leinonen
#           Lily Diaz
#           Janne Pietarila
#
# Code snippets used from Fle3 (fle3.uiah.fi),
# credits go to the Fle3 Team.
#
# This file is part of the ImaNote product.
#
# ImaNote is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# ImaNote is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ImaNote; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# Python
from PIL import Image, ImageDraw
from cStringIO import StringIO
import string
import os.path
import re
import time
# Zope
import DateTime
from AccessControl import ClassSecurityInfo
from Globals import InitializeClass, Persistent, package_home
from Acquisition import Implicit
import Acquisition
from AccessControl.Role import RoleManager
from OFS.Folder import Folder
from Products.PageTemplates.ZopePageTemplate import ZopePageTemplate
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
# Own
from common import *
from Annotation import Annotation, parseKeywords
import pdb
from AnnotationManager import AnnotationManager

MINIMUM_SIZE = 0


class TiledImage(Implicit, Persistent, RoleManager, Folder):
    """
    This is class for viewing large images (without annotations).
    """

    meta_type = 'TiledImage'
    manage_options=(Folder.manage_options)

    security = ClassSecurityInfo()
    security.declareObjectPublic()

    def __init__(self, my_id, magnificationStep, imageFile, title, \
        tileWidth=128, tileHeight=128, forceRgb=0):
      """Add a new map to this Made instance"""
      self.id = my_id
      self.title = title

      image = Image.open(StringIO(imageFile.read()))
      self._width, self._height = image.size

      # Do we convert even grayscale images to RGB?
      if forceRgb:
          self._mode = 'RGB'
          if image.mode != self._mode:
            image = image.convert(self._mode)
      else:
          if image.mode in ['1', 'L']: self._mode = 'L'
          else: self._mode = 'RGB'
      self._tile_width = tileWidth
      self._tile_height = tileHeight
      self._magnification_step = int(magnificationStep)
      
      # number of tiles in x and y directions
      self._n_tiles_x = [roundup(float(self._width) / float(tileWidth))]
      self._n_tiles_y = [roundup(float(self._height) / float(tileHeight))]
    
      # Create tiles by copying parts from the original image.
      # Tiles for other levels are created in manage_afterAdd().
      for y in range(self._n_tiles_y[0]): 
        for x in range(self._n_tiles_x[0]):
          tile = Image.new(self._mode, (tileWidth, tileHeight))
          tile.paste(
              image.crop(( x     * tileWidth,
                           y     * tileHeight,
                           min(self._width, (x+1) * tileWidth),
                           min(self._height,(y+1) * tileHeight))),
              (0,0))
          s = StringIO()
          tile.save(s, 'PNG')
          s.seek(0)
          self.addTile(0, x, y, s.read())

      del image
      # Create empty tile.
      image = Image.new(self._mode, (1,1))
      s.seek(0)
      image.save(s, 'PNG')
      s.seek(0)
      self.manage_addImage('tile_empty', s.read())

      del image
      #Create empty tile that has same size as the normal tiles
      image = Image.new(self._mode, (tileWidth,tileHeight))
      s.seek(0)
      image.save(s, 'PNG')
      s.seek(0)
      self.manage_addImage('tile_empty_with_dimensions', s.read())

      # Create tiles for other (down scaled) levels.
      self.downScale(0, self._n_tiles_x[0], self._n_tiles_y[0])
      self.setOverviewImageSize(3,3) # initialize overview image


    security.declareProtected(PERM_VIEW, 'index_html')
    def index_html(self):
      """Return map page"""
      return getattr(self, 'map.html')()

    security.declareProtected(PERM_VIEW, 'getTitle')
    def getTitle(self):
      """Get map title"""
      return self.title
    
    security.declarePrivate('manage_AfterAdd')
    def manage_afterAdd(self, item, container):
        """The rest of initialization code."""
        self.manage_addProperty('show_annotations', self.id, 'tokens')
        self.manage_addProperty('edit_annotations', self.id, 'tokens')
        #Store colors as hex for javascript
        self.manage_addProperty('color_1', 'ffffff', 'string')
        self.manage_addProperty('color_2', '000000', 'string')
        self.manage_addProperty('anno_min_size', '60', 'int')
        self.manage_addProperty('edit_disabled', '0', 'int')
        self.annotations._setObject(self.id, AnnotationManager(self.id))

    security.declarePrivate('addTile')
    def addTile(self, level, x, y, data):
      """Add one tile."""
      self.manage_addImage('tile_%s_%s_%s' % (level, x, y), data)

    security.declarePrivate('getTile')
    def getTile(self, level, x, y):
      return self._getOb('tile_%s_%s_%s' % (level, x, y)).data

    security.declareProtected(PERM_VIEW, 'getImageTile')
    def getImageTile(self, level, x, y):
      """Return one tile as image"""
      try:
        tileOb=self.getTile(level,x,y)
        tile=Image.open(StringIO(tileOb))
      except AttributeError:
        tile=Image.open(StringIO(self._getOb(\
            'tile_empty_with_dimensions').data))
      s = StringIO()
      tile.save(s, 'PNG')
      s.seek(0)
      if self.REQUEST and hasattr(self.REQUEST, 'RESPONSE'):
        self.REQUEST.RESPONSE.setHeader('content-type', 'image/png')
      return s.read()
         
    security.declarePrivate('downScale')
    def downScale(self, level, n_tiles_x, n_tiles_y):
        """Downscale current level (level) with n_tiles_x by n_tiles_y to
        create level+1. (Actually, in the end this calls itself
        recursively, so given original level (0) this creates all other
        levels 1,2,...n. """

        print level

        m = self._magnification_step

        width = roundup(self._width / pow(m, level))
        height = roundup(self._width / pow(m, level))

        n_tiles_x = roundup(n_tiles_x / float(m))
        n_tiles_y = roundup(n_tiles_y / float(m))

        self._n_tiles_x.append(n_tiles_x)
        self._n_tiles_y.append(n_tiles_y)

        s = StringIO()
        im = Image.new(self._mode, (m * self._tile_width,
                                     m * self._tile_height),
                       None)

        dummy = Image.new(self._mode, (self._tile_width,
                                        self._tile_height))

        for y in range(n_tiles_y):
            for x in range(n_tiles_x):

                # m = 2:        m = 3:
                # . .           . . .
                # . .           . . .
                #               . . .
                for yy in range(m):
                    for xx in range(m):
                #FIXME If tilesize is over ~150, we get 
                #TypeError: expected read buffer,
                #ImplicitAcquirerWrapper found
                        try:
                            im_tile = Image.open(StringIO(self.getTile( \
                                level, m*x+xx, m*y+yy)))
                        except AttributeError:
                            im_tile = dummy

                        im.paste(im_tile, (self._tile_width*xx,
                                           self._tile_height*yy))
                    
                scaled_im = im.resize((self._tile_width, self._tile_height),
                                      Image.ANTIALIAS)
                s.seek(0)
                scaled_im.save(s, 'PNG')
                s.seek(0)
                self.addTile(level+1, x, y, s)
                
        if width > 1 and height > 1:
            self.downScale(level + 1, n_tiles_x, n_tiles_y)

    security.declarePrivate('setOverviewImageSize')
    def setOverviewImageSize(self, max_width, max_height):
        """Set size of overview image so that is not bigger than max_width
        by max_height pixels. The aspect ratio of image is preserved. (If
        the aspect ratio of map is identical to max_width/max_height then
        size of overview will be exactly max_width by max_height pixels)."""

        overview_level = 0
        overview_width  = self._width
        overview_height = self._height
        magnification = self._magnification_step

        while ((overview_width  > magnification * max_width) or
               (overview_height > magnification * max_height)):
            overview_level += 1
            m = pow(self._magnification_step, overview_level)
            overview_width  = int(self._width  / m)
            overview_height = int(self._height / m)

        s = TiledImage.get_image(self, overview_level,
                           self._width / 2,
                           self._height / 2,
                           overview_width, overview_height, 'png')
        im = Image.open(StringIO(s))
        if im.mode != 'RGB':
            im = im.convert('RGB')
        im.thumbnail((max_width, max_height))

        s = StringIO()
        im.save(s, 'PNG')
        s.seek(0)
        self._overview_image = s.read()

        (self._overview_width, self._overview_height) = im.size
        (self._overview_max_width, self._overview_max_height) = \
                                    (max_width, max_height)

    security.declarePrivate('parametersFromRequest')
    def parametersFromRequest(self, ignore_click=0):
        """Return
           (level, center_x, center_y, width, height, format, click_x, click_y)
           tuple."""
        REQUEST=self.REQUEST

        (level, center_x, center_y, width, height, format, show_all) = \
                (3,               #
                 self._width/2,   #  default parameters
                 self._height/2,  #
                 640, 480, 'png', #
                 True)            #

        # Check what integer parameters are in REQUEST.
        for param in ['level', 'center_x', 'center_y', 'width', 'height',
                      'show_all']:
            if hasattr(REQUEST, param):
                exec('%s = int(REQUEST.%s)' % (param, param))

        # Our only string parameter.
        if hasattr(REQUEST, 'format'):
            format = REQUEST.format

        if not ignore_click and \
               hasattr(REQUEST, 'ov_point') and REQUEST.ov_point:
            # User clicked the overview map, calculate center_x and center_y
            # from that information.
            ov_point=REQUEST.ov_point
            if '?' in ov_point:
              ov_point=ov_point[1:]
            (ov_x, ov_y) = [float(x) for x in \
                            ov_point.split(',')]
            center_x = int((ov_x / self._overview_width ) * self._width)
            center_y = int((ov_y / self._overview_height) * self._height)
            (click_x, click_y)  = (width / 2, height / 2)
        else:
            if not ignore_click and \
                   hasattr(REQUEST, 'click_point') and REQUEST.click_point:
                # User has clicked the map itself.
                click_point=REQUEST.click_point
                if '?' in click_point:
                  click_point=click_point[1:]
                (click_x, click_y) = [int(x) for x in \
                    click_point.split(',')]
            else:
                (click_x, click_y)  = (width / 2, height / 2)

            m = int(pow(self._magnification_step, level))
            center_x += m * (click_x - width / 2)
            center_y += m * (click_y - height / 2)

        return (level, center_x, center_y, width, height, format,
                click_x, click_y, int(show_all))

    security.declareProtected(PERM_VIEW, 'get_overview_image')
    def get_overview_image(self, level, center_x, center_y, width, height,\
        useJavascript):
        """Return an overview image, where the current position in the
        actual image is indicated with a rectangle.

                     level: level of actual image
        center_x, center_y: position in the actual image
             width, height: size of "window" where actual image is displayed.
             """

        (level, center_x, center_y, width, height, useJavascript) = \
                [int(x) for x in (level, center_x, center_y, width, height,\
                                  useJavascript)]

        # Get image, on top on which we are going to draw rectangle
        im = Image.open(StringIO(self._overview_image))

        # Draw the rectangle
        if not useJavascript:
          draw = ImageDraw.Draw(im)
          m = pow(self._magnification_step, level)
          center_x = int(center_x *(float(self._overview_width)/self._width))
          center_y = int(center_y *(float(self._overview_height)/self._height))
          rectangle_width = int(
              (float(width)/self._width*m)*self._overview_width)
          rectangle_height = int(
              (float(height)/self._height*m)*self._overview_height)
          x1 = center_x - rectangle_width/2
          x2 = center_x + rectangle_width/2
          y1 = center_y - rectangle_height/2
          y2 = center_y + rectangle_height/2
          draw.rectangle((x1,y1,x2,y2), outline=(255,0,0))
        
         # Return image
          del draw
        s = StringIO()
        im.save(s, 'PNG')
        s.seek(0)
        self.REQUEST.RESPONSE.setHeader('content-type', 'image/png')
        return s.read()

    security.declarePrivate('get_im')
    def get_im(self, level, center_x, center_y, width, height):
        """Return an image (of size width*height) from given level on
           the map and location centered on (center_x, center_y)."""


        m = float(pow(self._magnification_step, level))
        start_x = int(center_x / m - (width / 2))
        start_y = int(center_y / m - (height / 2))

        im = Image.new(self._mode, (width, height))

        # Paste tiles to create the image.
        tile_y = start_y / self._tile_height
        y = - (start_y % self._tile_height)
        while y < height:
            tile_x = start_x / self._tile_width
            x = - (start_x % self._tile_width)
            while x < width:
                try:
                    tile = Image.open(StringIO(
                        self.getTile(level, tile_x, tile_y)))
                    im.paste(tile, (x, y))
                except:
                    pass
                tile_x += 1
                x += self._tile_width

            tile_y += 1
            y += self._tile_height

        return im
    security.declareProtected(PERM_VIEW, 'get_image')
    def get_image(self, level, center_x, center_y, width, height,
                  format='png', REQUEST=None):
        """Return one image. (Wrapper around get_im() method, this method
           converts data to JPEG or PNG and sets content-type in the header."""
        REQUEST=self.REQUEST

        (level, center_x, center_y, width, height) = \
                [int(x) for x in (level, center_x, center_y, width, height)]

        im = self.get_im(level, center_x, center_y, width, height)

        s = StringIO()
        if format == 'jpeg':
            im.save(s, 'JPEG', quality=50)
            s.seek(0)
            #we call self.REQUEST before object is added to ZODB so its a dummy
            #string
            if REQUEST and hasattr(REQUEST, 'RESPONSE'):
                REQUEST.RESPONSE.setHeader('content-type', 'image/jpeg')
        else:
            im.save(s, 'PNG')
            s.seek(0)
            if REQUEST and hasattr(REQUEST, 'RESPONSE'):
                REQUEST.RESPONSE.setHeader('content-type', 'image/png')
        return s.read()

    security.declareProtected(PERM_VIEW, 'get_format_url')
    def get_format_url(self, page_url, image_format, link_name):
        """Return HTML string (a link for changing image format)."""
        (level, center_x, center_y, width, height, format, dummy, dummy,\
            show_all) = self.parametersFromRequest()
        format = image_format
        return page_url + '?' + \
            string.join([str(x) + '=' + str(eval(x)) for x in \
            ('level', 'center_x', 'center_y', 'width',
              'height', 'format', 'show_all')], '&') + \
                  '&click_point='

        
    security.declareProtected(PERM_VIEW, 'get_zoom_url')
    def get_zoom_url(self, page_url, zoom_level):
        """Return HTML string (a link to change zoom level.)"""
        (level, center_x, center_y, width, height, format, dummy, dummy,\
            show_all) = self.parametersFromRequest()

        level = zoom_level
        return page_url + '?' + \
               string.join([str(x) + '=' + str(eval(x)) for x in \
               ('level', 'center_x', 'center_y', 'width',
                 'height', 'format', 'show_all')], '&') + \
                     '&click_point='

    security.declareProtected(PERM_VIEW, 'get_show_url')
    def get_show_url(self, page_url, show=''):
        """Return HTML string (a link to change whether own annos are show.)"""
        (level, center_x, center_y, width, height, format, dummy, dummy,\
            show_all) = self.parametersFromRequest()
        show_all=show
        return page_url + '?' + \
               string.join([str(x) + '=' + str(eval(x)) for x in \
               ('level', 'center_x', 'center_y', 'width',
                 'height', 'format', 'show_all')], '&') + \
                     '&click_point='


    security.declareProtected(PERM_VIEW, 'get_window_size_url')
    def get_window_size_url(self, page_url, window_width, window_height,
                            link_name):
        """Return a string of HTML <a>...</a> for selecting 'window' size."""
        (level, center_x, center_y, width, height, format, dummy, dummy,\
            show_all) = self.parametersFromRequest()
        width = window_width
        height = window_height
        
        return page_url + '?' + \
            string.join([str(x) + '=' + str(eval(x)) for x in \
            ('level', 'center_x', 'center_y', 'width',
              'height', 'format', 'show_all')], '&') + \
                  '&click_point='  

    security.declareProtected(PERM_VIEW, 'append_from_request')
    def append_from_request(self, url, ignore_click=0):
        """Append parameters from request to given url."""
        (level, center_x, center_y, width, height, format, click_x, click_y,\
            show_all) = self.parametersFromRequest(ignore_click)

        if '?' in url:
            char = '&'
        else:
            char = '?'
        
        retval = url + char + \
               string.join([str(x) + '=' + str(eval(x)) for x in \
                            ('level', 'center_x', 'center_y', 'width',
                             'height', 'format', 'click_x', 'click_y', 
                                 'show_all')], '&')
        return retval

    security.declareProtected(PERM_VIEW, 'append_simple_from_request')
    def append_simple_from_request(self, url):
        """Append parameters from request to given url (does not modify
           center_x and center_y unlike append_from_request())."""
        (level, center_x, center_y, width, height, format, click_x, click_y,\
            show_all) = self.parametersSimpleFromRequest()

        if '?' in url:
            char = '&'
        else:
            char = '?'
        
        retval = url + char + \
               string.join([str(x) + '=' + str(eval(x)) for x in \
                            ('level', 'center_x', 'center_y', 'width',
                             'height', 'format', 'click_x', 'click_y', \
                                 'show_all')], '&')
        return retval

    security.declareProtected(PERM_VIEW, 'get_image_url')
    def get_image_url(self, page_url, ignore_click=0):
        """Return URL string for map itself."""
        (level, center_x, center_y, width, height, format, click_x, click_y,\
            show_all) = self.parametersFromRequest(ignore_click)

        if '?' in page_url:
            char = '&'
        else:
            char = '?'

        return page_url + char + \
               string.join([str(x) + '=' + str(eval(x)) for x in \
                            ('level', 'center_x', 'center_y', 'width',
                             'height', 'format', 'show_all')], '&') + '&click_point='  
        
    security.declareProtected(PERM_VIEW, 'get_image_map_url')
    def get_image_map_url(self, ignore_click=0):
      """Return URL string for map image."""

      (level, center_x, center_y, width, height, format, click_x, click_y,\
          show_all) = self.parametersFromRequest(ignore_click)   
      return  'get_image?' + string.join([str(x) + '=' + str(eval(x)) for x in \
                         ('level', 'center_x', 'center_y', 'width',
                          'height', 'format', 'show_all')], '&')
        
    security.declareProtected(PERM_VIEW, 'get_overview_url')
    def get_overview_url(self, page_url, ignore_click=0):
        """Return URL string for overview map."""
        (level, center_x, center_y, width, height, format, dummy, dummy,\
            show_all) = self.parametersFromRequest(ignore_click)

        return page_url + '?' + \
               string.join([str(x) + '=' + str(eval(x)) for x in \
                            ('level', 'center_x', 'center_y', 'width',
                             'height', 'format', 'show_all')], '&') + '&ov_point='

    security.declareProtected(PERM_VIEW, 'get_overview_image_url')
    def get_overview_image_url(self, overview_max_width, overview_max_height,
                               ignore_click=0, useJavascript=False):
      """Return URL string for overview image."""
      (level, center_x, center_y, width, height, format, dummy, dummy,\
          show_all) = self.parametersFromRequest(ignore_click)
      overview_max_width  = int(overview_max_width)
      overview_max_height = int(overview_max_height)
      
      if ((overview_max_width  != self._overview_max_width) or
          (overview_max_height != self._overview_max_height)):
        self.setOverviewImageSize(overview_max_width, overview_max_height)
      
      overview_width = self._overview_width
      overview_height = self._overview_height
      
      useJavascript=int(useJavascript)
      return  'get_overview_image?' + \
          'overview_width=%s&overview_height=%s&' \
          % (overview_width, overview_height) + \
          string.join([str(x) + '=' + str(eval(x)) for x in \
          ('level', 'center_x', 'center_y', 'width',
            'height', 'format', 'show_all', 'useJavascript')], '&')
            
#Some getters for the javascript code
    security.declareProtected(PERM_VIEW, 'getOverviewWidth')
    def getOverviewWidth(self):
      """Return overview image width"""
      return self._overview_width
    
    security.declareProtected(PERM_VIEW, 'getOverviewHeight')      
    def getOverviewHeight(self):
      """Return overview image height"""
      return self._overview_height

    security.declareProtected(PERM_VIEW, 'getImageWidth')
    def getImageWidth(self):
      """Return image width"""
      return self._width
      
    security.declareProtected(PERM_VIEW, 'getImageHeight')
    def getImageHeight(self):
      """Return image height"""
      return self._height
    
    security.declareProtected(PERM_VIEW, 'getTileWidth')
    def getTileWidth(self):
      """Return used tile width"""
      return self._tile_width
      
    security.declareProtected(PERM_VIEW, 'getTileHeight')
    def getTileHeight(self):
      """Return used tile height"""
      return self._tile_height
      
    security.declareProtected(PERM_VIEW, 'getMagnification')
    def getMagnification(self):
      """Return magnification factor"""
      (level, center_x, center_y, width, height, format, click_x, click_y,\
          show_all) = self.parametersFromRequest()
      return int(pow(self._magnification_step, level))

    security.declareProtected(PERM_VIEW, 'getLevel')
    def getLevel(self):
      """Return magnification level"""
      (level, center_x, center_y, width, height, format, click_x, click_y,\
          show_all) = self.parametersFromRequest()
      return level

    security.declareProtected(PERM_VIEW, 'getCenterX')
    def getCenterX(self):
      (level, center_x, center_y, width, height, format, click_x, click_y,\
          show_all) = self.parametersFromRequest()
      return center_x

    security.declareProtected(PERM_VIEW, 'getCenterY')
    def getCenterY(self):
      (level, center_x, center_y, width, height, format, click_x, click_y,\
          show_all) = self.parametersFromRequest()
      return center_y
      
      

class AnnotatedTiledImage(TiledImage):
    """
    This class adds annotations to TiledImage.
    """

    meta_type = 'AnnotatedTiledImage'
    security = ClassSecurityInfo()
    security.declareObjectPublic()
    def __init__(self, my_id, magnificationStep, imageFile, imageTitle, \
        tileWidth=128, tileHeight=128, forceRgb=0):
        TiledImage.__init__(self, my_id, magnificationStep, imageFile,\
            imageTitle, tileWidth, tileHeight, forceRgb)

        # FIXME: This has to be replaced with some clever data
        # FIXME: structure so that we can find annotations in
        # FIXME: given area fast...

    security.declarePrivate('manage_AfterAdd')
    def manage_afterAdd(self, item, container):
        TiledImage.manage_afterAdd(self, item, container)

    security.declarePrivate('addAnnotation')
    def addAnnotation(self, annotation):
      for i in self.edit_annotations:
        annos=getattr(self.annotations, i)
        annos.addAnnotation(annotation) 

    # world coordinates: normal pixel coordinates on the original image (i.e.
    # image on level 0)
    # window coordinates: I call window the part of the image that is
    # displayed on the web page. Window coordinates are normal pixel
    # coordinates on that one.
    security.declarePrivate('worldToWindow')
    def worldToWindow(self, x_world, y_world,
                        level, center_x, center_y, width, height):
        """Converts (x_world, y_world) from world coordinates to
        window coordinates."""

        m = float(pow(self._magnification_step, level))
        x_window = int((  width / 2) - center_x/m + x_world/m)
        y_window = int(( height / 2) - center_y/m + y_world/m)
        return (x_window, y_window)

    security.declarePrivate('windowToWorld')
    def windowToWorld(self, x_window, y_window,
                        level, center_x, center_y, width, height):
        """Converts (x_window, y_window) from window coordinates to
        world coordinates."""

        m = float(pow(self._magnification_step, level))
        x_world = int(m*x_window -  m*width/2 + center_x)
        y_world = int(m*y_window - m*height/2 + center_y)
        return (x_world, y_world)

    def getClickedAnnotation(self):
        if not hasattr(self.REQUEST, 'click_point') or not self.REQUEST.click_point:
          # The user did not click the image at all!
          return None
        (level, center_x, center_y, width, height, format, click_x, click_y,\
            show_all) = self.parametersSimpleFromRequest()
        (x,y) = self.windowToWorld(click_x, click_y, level,
                                     center_x, center_y, width, height)
        if hasattr(self.REQUEST, 'a_id'): 
          a_id=self.REQUEST.a_id
          t = [self.getAnnotation(a_id)]
        else:
          t=self.getAnnotationsForPoint(x,y, show_all)
        if t:
          for annotation in t:
            # Does the annotation cover the whole window...?
            (left, top, right, bottom) = annotation.getCoordinates()
            (left, top) = self.worldToWindow(left, top,
                                               level, center_x, center_y,
                                               width, height)
            (right, bottom) = self.worldToWindow(right, bottom,
                                                   level, center_x, center_y,
                                                   width, height)
            if (left < 0 and right >= width-1 and \
                top < 0 and bottom >= height -1):
                # ... if so, do not react to click.
                t.remove(annotation)

            if (right-left) * (bottom-top) < MINIMUM_SIZE:
                t.remove(annotation)
          return t
        return None

    security.declarePrivate('getAnnotationsForPoint')
    def getAnnotationsForPoint(self, x, y, show_all=True):
        """Returns an Annotation object that covers point (x,y) (in world
        coordinates). If there is not any object covering the point, returns
        None. If there are several objects covering the given point, the one
        having the smallest area is returned."""
        annos = []
        for (i,a) in self.getAnnotationsList(show_all).items():
            if a.doesCover(x,y):
                annos.append(a)

        if len(annos) == 0:
            return None
        else:
          return annos

    security.declareProtected(PERM_EDIT, 'get_image_with_rect_url')
    def get_image_with_rect_url(self):
        """Return URL string (img url)."""
        (level, center_x, center_y, width, height, format, click_x, click_y,\
            show_all) = self.parametersFromRequest()
        if hasattr(self.REQUEST, 'coord_x_1'):
          coord_x_1=int(self.REQUEST.coord_x_1)
        if hasattr(self.REQUEST, 'coord_y_1'):
          coord_y_1=int(self.REQUEST.coord_y_1)
        if hasattr(self.REQUEST, 'coord_x_2'):
          coord_x_2=int(self.REQUEST.coord_x_2)
        if hasattr(self.REQUEST, 'coord_y_2'):
          coord_y_2=int(self.REQUEST.coord_y_2)

        return self.absolute_url() + '/get_image_with_rect?' + \
            string.join([str(x) + '=' + str(eval(x)) for x in \
                         ('level', 'center_x', 'center_y', 'width',
                          'height', 'format', 'show_all',
                          'coord_x_1', 'coord_y_1',
                          'coord_x_2', 'coord_y_2')], '&') +'&click_point='

    security.declareProtected(PERM_EDIT, 'get_image_with_rect')
    def get_image_with_rect(self,  level, center_x, center_y, width, height,
                            coord_x_1, coord_y_1, coord_x_2, coord_y_2,
                            format='png', REQUEST=None):
        """Return an image (of size width*height) from given level on the map
           and location centered on (center_x, center_y) with rectangle in
           location (coord_x_1, coord_y_1) - (coord_x_2, coord_y_2)."""
        REQUEST=self.REQUEST

        (level, center_x, center_y, width, height, coord_x_1, coord_y_1,
         coord_x_2, coord_y_2) = \
         [int(x) for x in (level, center_x, center_y, width, height,
                           coord_x_1, coord_y_1, coord_x_2, coord_y_2)]

        im = self.get_im(level, center_x, center_y, width, height)
        draw = ImageDraw.Draw(im)

        (x1, y1) = self.worldToWindow(coord_x_1, coord_y_1,
                                        level, center_x, center_y,
                                        width, height)
        (x2, y2) = self.worldToWindow(coord_x_2, coord_y_2,
                                        level, center_x, center_y,
                                        width, height)

        colour1=(int('0x'+self.color_1[0:2],16),\
            int('0x'+self.color_1[2:4],16),\
            int('0x'+self.color_1[4:6],16))
        colour2=(int('0x'+self.color_2[0:2],16),\
            int('0x'+self.color_2[2:4],16),\
            int('0x'+self.color_2[4:6],16))
        draw.rectangle((x1, y1, x2+1, y2+1), outline=colour1)
        draw.rectangle((x1+1, y1+1, x2-1, y2-1), outline=colour2)
        del draw
        s = StringIO()
        im.save(s, 'PNG')
        s.seek(0)
        self.REQUEST.RESPONSE.setHeader('content-type', 'image/png')
        return s.read()

    security.declareProtected(PERM_EDIT, 'get_image_with_haircross_link_url')
    def get_image_with_haircross_link_url(
        self, page_url):
        """Return URL string (map link...)"""
        (level, center_x, center_y, width, height, format, click_x, click_y,\
            show_all) = self.parametersFromRequest()

        if '?' in page_url:
            char = '&'
        else:
            char = '?'
        if hasattr(self.REQUEST, 'coord_x'):
          coord_x=int(self.REQUEST.coord_x)
        if hasattr(self.REQUEST, 'coord_y'):
          coord_y=int(self.REQUEST.coord_y)

        return page_url + char + \
               string.join([str(x) + '=' + str(eval(x)) for x in \
                            ('level', 'center_x', 'center_y', 'width',
                             'height', 'format', 'coord_x', 'coord_y', \
                                 'show_all')],
                           '&') + '&click_point='

    security.declareProtected(PERM_EDIT, 'get_image_with_haircross_url')
    def get_image_with_haircross_url(self):
      """Return URL string (map image)"""
      (level, center_x, center_y, width, height, format, click_x, click_y,\
          show_all) = self.parametersFromRequest()
      if hasattr(self.REQUEST, 'coord_x'):
        coord_x=int(self.REQUEST.coord_x)
      if hasattr(self.REQUEST, 'coord_y'):
        coord_y=int(self.REQUEST.coord_y)
          
      return self.absolute_url() + '/get_image_with_haircross?' + \
          string.join([str(x) + '=' + str(eval(x)) for x in \
          ('level', 'center_x', 'center_y', 'width',
            'height', 'format',
            'coord_x', 'coord_y', 'show_all')], '&')
          
    security.declareProtected(PERM_EDIT, 'get_image_with_haircross')
    def get_image_with_haircross(self,  level, center_x, center_y,
                                 width, height,
                                 coord_x, coord_y,
                                 format='png', REQUEST=None):
        """Return an image (of size width*height) from given level on the map
           and location centered on (center_x, center_y) with haircross
           centered on (coord_x, coord_y)."""
        REQUEST=self.REQUEST
        (level, center_x, center_y, width, height, coord_x, coord_y) = \
                [int(x) for x in (level, center_x, center_y, width, height,
                                  coord_x, coord_y)]
        
        im = self.get_im(level, center_x, center_y, width, height)
        draw = ImageDraw.Draw(im)

        (x, y) = self.worldToWindow(coord_x, coord_y,
                                      level, center_x, center_y,
                                      width, height)

        draw.line((0, y, width, y),  fill=(0,255,0))
        draw.line((x, 0, x, height), fill=(0,255,0))

        del draw
        s = StringIO()
        im.save(s, 'PNG')
        s.seek(0)
        self.REQUEST.RESPONSE.setHeader('content-type', 'image/png')
        return s.read()

    security.declareProtected(PERM_VIEW, 'get_image')
    def get_image(self,
                  level, center_x, center_y, width, height,
                  format='png', REQUEST=None):
        """Overdrive get_image() in TiledImage.
        Return image with annotations."""
        REQUEST=self.REQUEST
        (level, center_x, center_y, width, height) = \
                [int(x) for x in (level, center_x, center_y, width, height)]
        
        im = self.get_im(level, center_x, center_y, width, height)
        draw = ImageDraw.Draw(im)

        for a in self.getAnnotationsList().values():
            (x1, y1, x2, y2) = a.getCoordinates()
            (x1, y1) = self.worldToWindow(x1, y1,
                                            level, center_x, center_y,
                                            width, height)
            (x2, y2) = self.worldToWindow(x2, y2,
                                            level, center_x, center_y,
                                            width, height)
            if (x2-x1) * (y2- y1) >= MINIMUM_SIZE:
                if a.getIcon():
                  im.paste(a.getIcon, (x1,y1))
                colour1=(int('0x'+self.color_1[0:2], 16),\
                    int('0x'+self.color_1[2:4], 16),\
                    int('0x'+self.color_1[4:6], 16))
                colour2=(int('0x'+self.color_2[0:2], 16),\
                    int('0x'+self.color_2[2:4], 16),\
                    int('0x'+self.color_2[4:6],16))
                draw.rectangle((x1, y1, x2+1, y2+1), outline=colour1)
                draw.rectangle((x1+1, y1+1, x2, y2), outline=colour2)
        del draw

        s = StringIO()
        if format == 'jpeg':
            im.save(s, 'JPEG', quality=50)
            s.seek(0)
            if REQUEST:
                REQUEST.RESPONSE.setHeader('content-type', 'image/jpeg')
        else:
            im.save(s, 'PNG')
            s.seek(0)
            if REQUEST:
                REQUEST.RESPONSE.setHeader('content-type', 'image/png')
        return s.read()

    security.declareProtected(PERM_EDIT, 'removeAnnotation')
    def removeAnnotation(self, a_id):
        """Remove given annotation by id."""
        REQUEST=self.REQUEST
        for i in self.edit_annotations:
          getattr(self.annotations, i).removeAnnotation(a_id)
        
        (level, center_x, center_y, width, height, format,
         click_x, click_y, show_all) = self.parametersSimpleFromRequest()
        REQUEST.RESPONSE.redirect(
            REQUEST.next + '?' +
            string.join([str(x) + '=' + str(eval(x)) for x in \
                         ('level', 'center_x', 'center_y', 'width',
                          'height', 'format', 'show_all')], '&'))

    security.declareProtected(PERM_VIEW, 'getAnnotation')
    def getAnnotation(self, a_id):
      """Retrieve an annotation with the given id"""
      for i in self.show_annotations:
        annos=getattr(self.annotations, i)
        for (id, anno) in annos.getAnnotations():
          if a_id==id: return anno
      raise 'AnnotationError', 'Annotation %s not found' %a_id
      
    security.declareProtected(PERM_VIEW, 'getAnnotations')
    def getAnnotations(self):
      """Retrieve annotations as list sorted by date"""
      values=self.getAnnotationsList().values()
      def dateSort(x, y):
        return int(y.getDate()-x.getDate())
      values.sort(dateSort)
      return values
    
    security.declareProtected(PERM_VIEW, 'getAnnotationsList')
    def getAnnotationsList(self, show_all=''):
      """Get all annotations attached to this image"""
      if show_all=='':
        show_all=True
        if hasattr(self.REQUEST, 'show_all') and self.REQUEST.show_all=='0':
          show_all=False
      retList={}
      for i in self.show_annotations:
        annos=getattr(self.annotations, i)
        for (id, anno) in annos.getAnnotations():
          if (not show_all) and not anno.mayEdit(): continue
          retList[id]=anno
      return retList
    
    security.declarePrivate('editAnnotation')
    def editAnnotation(self, id, coords='', title='', author='', url='',\
        urlName='', text='', description='', public='', keywords='', \
        urlAuthor=''):
      """Edit annotation data"""
      for i in self.edit_annotations:
        annos=getattr(self.annotations, i)
        for (a_id, anno) in annos.getAnnotations():
          if id==a_id and anno.mayEdit():
            anno.editData(coords, title, author, url, urlName, text,\
                description, public, keywords, urlAuthor)
            return
      raise 'AnnotationError','Annotation %s not found or no rights to edit' %id
      
    security.declareProtected(PERM_EDIT, 'updateAnnotationPosition')
    def updateAnnotationPosition(self, a_id, newX1, newY1, width=0, height=0, level=0, absolute='true'):
      """Update position of annotation - if absolute is not set, use relative pixel value"""
      anno=None
      for i in self.edit_annotations:
        annos=getattr(self.annotations, i)
        for (id, anno) in annos.getAnnotations():
          if id==a_id:
            break
      if anno==None:
        raise 'AnnotationError', 'Annotation %s not found' %a_id

      (newX1, newY1, width, height, level)=\
          [int(x) for x in (newX1, newY1, width, height, level)]
      (newX1, newY1, width, height)=\
          [i*pow(self._magnification_step, level) \
          for i in (newX1, newY1, width, height)]
      if absolute=='true':
        anno.moveTo(newX1,newY1)
      else:
        anno.move(newX1,newY1)
      if width!=0 and height !=0:
        anno.setDimensions(width, height)

    security.declareProtected(PERM_VIEW, 'isAnnotationWritable')
    def isAnnotationWritable(self, id):
      for i in self.edit_annotations:
        annos=getattr(self.annotations, i)
        anno=annos.getAnnotationById(id)
        if anno:
          return int(anno.mayEdit())
      return 0

    security.declareProtected(PERM_VIEW, 'getAnnotationsForTile')
    def getAnnotationsForTiles(self, x, y):
      """Retrieve annotations that are located in given tile"""
      values=self.getAnnotationsList(show_all).values()
      (level, center_x, center_y, width, height, format, click_x, click_y,\
          show_all) = self.parametersSimpleFromRequest()
      m = pow(self._magnification_step, level)
      #TODO Not ready yet
      for i in values:
        pass
        
        

    security.declareProtected(PERM_VIEW, 'getAnnotationsForArea')
    def getAnnotationsForArea(self, getAll=False):
      """Retrieve annotations located in the current area"""
      (level, center_x, center_y, width, height, format, click_x, click_y,\
          show_all) = self.parametersSimpleFromRequest()
      (x,y)=self.windowToWorld(click_x, click_y, level, center_x, center_y,\
          width, height)
      m = pow(self._magnification_step, level)
      #Window area in world coordinates
      start_x=x-m*(width/2)
      start_y=y-m*(height/2)
      end_x=x+m*(width/2)
      end_y=y+m*(height/2)
      retVal=[]
      anno_min_size=self.anno_min_size*m*m #Scale min_size to current level
      values=self.getAnnotationsList(show_all).values()
      for i in values:
        if m!=1 and i.getSize()<anno_min_size:
          continue
        if not getAll:
        #Check that annotation is somewhere in the window
          (x1, y1, x2, y2)=i.getCoordinates()
          if not (((x1>start_x and x1<end_x) or \
              (x2>start_x and x2<end_x)) and \
              ((y1>start_y and y1<end_y) or\
              (y2>start_y and y2<end_y))):
            continue
        retVal.append(i)
      return retVal
          
          
          
    security.declareProtected(PERM_VIEW, 'getLinkToAnnotation')
    def getLinkToAnnotation(self, a_id):
      """Retrieve an annotation with the given id"""
      (level, center_x, center_y, width, height, format,
          click_x, click_y, show_all) = self.parametersSimpleFromRequest()
      pos=self.getAnnotation(a_id).getPosition\
          (width,height,self._magnification_step)
      wi=width/2
      hei=height/2
      return '%s?level=%s&center_x=%s&center_y=%s&width=%s&height=%s&show_all=%s&format=png&click_point=%s,%s&a_id=%s' % (self.absolute_url(), pos[2], pos[0],\
          pos[1], width, height, show_all, wi, hei, a_id)

    security.declareProtected(PERM_VIEW, 'locateAnnotation')
    def locateAnnotation(self, id):
      """Locate annotation on image and redirect"""
      self.REQUEST.RESPONSE.redirect(self.getLinkToAnnotation(id))


    security.declareProtected(PERM_VIEW, 'search_annotation_handler')
    def search_annotation_handler(self, cancel=None, submit=None):
      """Handler to handle annotation searching form"""
      (level, center_x, center_y, width, height, format,
        click_x, click_y, show_all) = self.parametersSimpleFromRequest()
      request=self.REQUEST
      if submit:
        kWords=parseKeywords(request['getKeywords'])
        request.set('getKeywords', kWords)
        
        return getattr(self, 'results.html')(self, self.REQUEST)
      else:
        self.REQUEST.next='index_html'
        self.REQUEST.RESPONSE.redirect(
            self.REQUEST.next + '?' +
            string.join([str(x) + '=' + str(eval(x)) for x in \
                ('level', 'center_x', 'center_y', 'width',
                  'height', 'format', 'show_all')], '&'))

    security.declareProtected(PERM_EDIT, 'add_annotation_handler_step_1')
    def add_annotation_handler_step_1(self):
        """Form handler for adding a new annotation (step 1/3)."""
        REQUEST=self.REQUEST
        (level, center_x, center_y, width, height, format,
         click_x, click_y, show_all) = self.parametersSimpleFromRequest()
        a_id=None
        if hasattr(REQUEST, 'a_id') and REQUEST.a_id!='None':
          a_id=REQUEST.a_id
        
        m = pow(self._magnification_step, level)

        #                     upper left
        #              ____ corner of window ___
        #             /                         \
        coord_x = int(center_x - m * ( width / 2) + m * click_x)
        coord_y = int(center_y - m * (height / 2) + m * click_y)

        REQUEST.RESPONSE.redirect(
            REQUEST.next + '?' +
            string.join([str(x) + '=' + str(eval(x)) for x in \
                         ('level', 'center_x', 'center_y', 'width',
                          'height', 'format', 'coord_x', 'coord_y', 'show_all',
                          'a_id')], '&'))


    security.declareProtected(PERM_EDIT, 'add_annotation_handler_step_2')
    def add_annotation_handler_step_2(self):
        """Form handler for adding a new annotation (step 2/3)."""
        REQUEST=self.REQUEST
        (level, center_x, center_y, width, height, format,
         click_x, click_y, show_all) = self.parametersSimpleFromRequest()
        a_id=None
        if hasattr(REQUEST, 'a_id') and REQUEST.a_id!='None':
          a_id=REQUEST.a_id
        
        m = pow(self._magnification_step, level)

        coord_x_1 = int(REQUEST.coord_x)
        coord_y_1 = int(REQUEST.coord_y)
        coord_x_2 = int(center_x - m * ( width / 2) + m * click_x)
        coord_y_2 = int(center_y - m * (height / 2) + m * click_y)
        #               \                         /
        #                \___    upper left    __/
        #                     corner of window 

        if coord_x_1 > coord_x_2:
            (coord_x_1, coord_x_2) = (coord_x_2, coord_x_1)
        if coord_y_1 > coord_y_2:
            (coord_y_1, coord_y_2) = (coord_y_2, coord_y_1)
        if a_id: #We are moving an existing annotation
          self.editAnnotation(a_id, coords=\
              (coord_x_1,coord_y_1,coord_x_2,coord_y_2))
          REQUEST.next='index_html'
          REQUEST.RESPONSE.redirect(
              REQUEST.next + '?' +
              string.join([str(x) + '=' + str(eval(x)) for x in \
                           ('level', 'center_x', 'center_y', 'width',
                            'height', 'format', 'show_all')], '&'))
        else:
          REQUEST.RESPONSE.redirect(
              REQUEST.next + '?' +
              string.join([str(x) + '=' + str(eval(x)) for x in \
                  ('level', 'center_x', 'center_y', 'width',
                    'height', 'format',
                    'coord_x_1', 'coord_y_1',
                    'coord_x_2', 'coord_y_2', 'show_all')], '&'))

    security.declareProtected(PERM_EDIT, 'add_annotation_handler_step_3')
    def add_annotation_handler_step_3(self, 
                                      cancel='', # submit buttons
                                      ok=''      #
                                      ):
        """Form handler for adding a new annotation (step 3/3)."""
        REQUEST=self.REQUEST

        if ok and not self.edit_disabled:
            coord_x_1 = int(REQUEST.coord_x_1)
            coord_y_1 = int(REQUEST.coord_y_1)
            coord_x_2 = int(REQUEST.coord_x_2)
            coord_y_2 = int(REQUEST.coord_y_2)
            left = min(coord_x_1, coord_x_2)
            right = max(coord_x_1, coord_x_2)
            top = min(coord_y_1, coord_y_2)
            bottom = max(coord_y_1, coord_y_2)
            id=self.generateId()
            self.addAnnotation(Annotation(left,top,right,bottom,id))
            url=""
            public=hasattr(REQUEST, 'annotation_public') and \
                REQUEST.annotation_public
            if not REQUEST.annotation_url=='http://':
              url=REQUEST.annotation_url
            self.editAnnotation(id, url=url, \
                author=self.getUser(),\
                text=REQUEST.annotation_text,\
                description=REQUEST.annotation_description,\
                urlName=REQUEST.annotation_urlName,\
                title=REQUEST.annotation_title, \
                keywords=REQUEST.annotation_keywords, \
                public=public, urlAuthor=REQUEST.annotation_urlAuthor
                )
            
            print 'added Annotation'
        elif cancel:
            pass # do nothing
        else:
            raise 'Unknown submit button.'

        (level, center_x, center_y, width, height, format,
         click_x, click_y, show_all) = self.parametersSimpleFromRequest()
        
        REQUEST.RESPONSE.redirect(
            REQUEST.next + '?' +
            string.join([str(x) + '=' + str(eval(x)) for x in \
                         ('level', 'center_x', 'center_y', 'width',
                          'height', 'format', 'show_all')], '&'))
        

    security.declareProtected(PERM_EDIT, 'edit_annotation_handler')
    def edit_annotation_handler(self,
                                cancel='', # submit buttons
                                ok=''      #
                                ):
        """Edit existing annotation."""
        REQUEST=self.REQUEST
        if ok and not self.edit_disabled:
          for i in ['url', 'urlName','text','description','title', 'keywords', 'urlAuthor']:
            exec('%s=REQUEST.annotation_%s' % (i,i))
            exec ('if %s==\'\': %s=None' % (i,i))
          public=hasattr(REQUEST, 'annotation_public') and REQUEST.annotation_public
          if REQUEST.annotation_url=='http://':
            url=None
          self.editAnnotation(id=REQUEST.a_id, url=url,
              text=text, description=description, urlName=urlName, title=title,
              public=public, keywords=keywords, urlAuthor=urlAuthor)
        elif cancel:
            pass
        else:
            raise 'Unknown submit button.'
            
        (level, center_x, center_y, width, height, format,
         click_x, click_y, show_all) = self.parametersSimpleFromRequest()

        REQUEST.RESPONSE.redirect(
            REQUEST.next + '?' +
            string.join([str(x) + '=' + str(eval(x)) for x in \
                         ('level', 'center_x', 'center_y', 'width',
                          'height', 'format', 'show_all')], '&'))
        
      

    security.declarePrivate('parametersSimpleFromRequest')
    def parametersSimpleFromRequest(self):
        """Return
           (level, center_x, center_y, width, height, format, click_x, click_y)
           tuple."""
        REQUEST=self.REQUEST
        (level, center_x, center_y, width, height, format, show_all) = \
                (3,               #
                 self._width/2,   #  default parameters
                 self._height/2,  #
                 640, 480, 'png', #
                 True )           #

        # Check what integer parameters are in REQUEST.
        for param in ['level', 'center_x', 'center_y', 'width', 'height',
                      'show_all']:
            if hasattr(REQUEST, param):
                exec('%s = int(REQUEST.%s)' % (param, param))

        # Our only string parameter.
        if hasattr(REQUEST, 'format'):
            format = REQUEST.format

        if hasattr(REQUEST, 'ov_point') and REQUEST.ov_point:
            # User clicked the overview map, calculate center_x and center_y
            # from that information.
            ov_point=REQUEST.ov_point
            if '?' in ov_point:
              ov_point=ov_point[1:]
            (ov_x, ov_y) = [float(x) for x in \
                            ov_point.split(',')]
            center_x = int((ov_x / self._overview_width ) * self._width)
            center_y = int((ov_y / self._overview_height) * self._height)
            (click_x, click_y)  = (width / 2, height / 2)

        if hasattr(REQUEST, 'click_point') and REQUEST.click_point:
          click_point=REQUEST.click_point
          if '?' in click_point:
            click_point=click_point[1:]
          (click_x, click_y) = [int(x) for x in \
                                click_point.split(',')]
        else :
            (click_x, click_y)  = (width / 2, height / 2)

        return (level, center_x, center_y, width, height, format,
                click_x, click_y, int(show_all))

InitializeClass(TiledImage)
InitializeClass(AnnotatedTiledImage)
